# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
### BEGIN LICENSE
# Copyright (C) 2012 Phoenix87 <phoenix1987@gmail.com>
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE

import gettext
from gettext import gettext as _
gettext.textdomain('gtumbler')

import gobject
import gtk
import logging
logger = logging.getLogger('gtumbler')

from gtumbler_lib import Window
from gtumbler_lib import pdf
from gtumbler_lib.command import Command

from gtumbler.AboutGtumblerDialog		import AboutGtumblerDialog
from gtumbler.PreferencesGtumblerDialog  import PreferencesGtumblerDialog
from gtumbler.ProgressDialog			 import ProgressDialog

import gtumbler.shell as shell
import gtumbler.mimes as mimes
import os
import session
import threading
import time
import tempfile
import cPickle

Session = session.Session()

### BEGIN THREAD CLASSES

class ProgressThread(threading.Thread):
	def __init__(self, parent):
		self.parent = parent # Parent progress dialog
		self._stop = threading.Event()
		threading.Thread.__init__(self)

	def stop(self):
		self._stop.set()
		
	def has_stopped(self):
		return self._stop.isSet()

class ConcatenateThread(ProgressThread):
	def __init__(self, parent, data):
		self.main = data
		ProgressThread.__init__(self, parent)

	def cat2(self, a, b, c):
		pdf_out = pdf.PdfFileWriter()
		with file(a, 'rb') as fin1, file(b, 'rb') as fin2, file(c, 'wb') as fout:
			pdf_in1 = pdf.PdfFileReader(fin1)
			pdf_in2 = pdf.PdfFileReader(fin2)
			
			self.parent.set_steps(pdf_in1.getNumPages() + pdf_in2.getNumPages())

			gtk.gdk.threads_enter()
			self.parent.set_current_task("Concatenating documents...")
			gtk.gdk.threads_leave()
			
			for page in pdf_in1.pages:
				pdf_out.addPage(page)
				gtk.gdk.threads_enter()
				self.parent.stepit()
				gtk.gdk.threads_leave()
				if self.has_stopped(): return

			for page in pdf_in2.pages:
				pdf_out.addPage(page)
				gtk.gdk.threads_enter()
				self.parent.stepit()
				gtk.gdk.threads_leave()
				if self.has_stopped(): return
			
			gtk.gdk.threads_enter()
			self.parent.set_current_task("Writing data to disk...")
			gtk.gdk.threads_leave()

			gtk.gdk.threads_enter()
			self.parent.spinner(True)
			gtk.gdk.threads_leave()
			
			pdf_out.write(fout)

			gtk.gdk.threads_enter()
			self.parent.set_current_task("Concatenating documents...")
			gtk.gdk.threads_leave()

			gtk.gdk.threads_enter()
			self.parent.spinner(False)
			gtk.gdk.threads_leave()

	#~ def catn(self, p, tmp):
		#~ if len(p) == 1: return p[0]
#~ 
		#~ if len(p) % 2:
			#~ odd = tmp.new_file('.pdf')
			#~ self.cat2(p[0], p[1], odd)
			#~ p[1] = odd
			#~ p = p[1:]
#~ 
		#~ i = 0
		#~ fr = self.parent.get_fraction()
		#~ n = len(p) / 2
		#~ while i < n:
			#~ new_file = tmp.new_file('.pdf')
			#~ self.cat2(p[2 * i], p[2 * i + 1], new_file)
			#~ p[i] = new_file
			#~ #gtk.gdk.threads_enter()
			#~ #self.parent.set_fraction(fr + (1-fr)/n * (i+1))
			#~ #gtk.gdk.threads_leave()
			#~ i += 1
#~ 
		#~ return self.catn(p[:len(p) / 2] , tmp)
		
	def catn(self, p, tmp):
		n = len(p)
		while n > 1:
			i = r = n % 2
			n = n >> 1
			fr = self.parent.get_fraction()
			while i < n + r:
				new_file = tmp.new_file('.pdf')
				self.cat2(p[2 * i - r], p[2 * i + 1 - r], new_file)
				p[i] = new_file
				i += 1
			n += r
		return p[0]

	def run(self):
		it = self.main.ui.ls_documents.get_iter_first()
		pdf_out = pdf.PdfFileWriter()
		
		tmp = Session.temp_new('cat')
		p = []
		gtk.gdk.threads_enter()
		self.parent.set_current_task("Scanning document list...")
		self.parent.spinner(True)
		gtk.gdk.threads_leave()

		while it:
			fname, path, mime = self.main.ui.ls_documents.get(it, 1, 2, 3)
			
			old_file = "%s/%s" % (path, fname)
			if mime in mimes.CONVERT_MIMES and self.main.ui.chk_ps_to_pdf.get_active(): # Must be converted
				new_file = tmp.new_file('.pdf')
				gtk.gdk.threads_enter()
				self.parent.set_current_task('Converting "%s" from PostScript to PDF...' % old_file)
				self.parent.spinner(True)
				gtk.gdk.threads_leave()
				
				shell.ps2pdf(old_file, new_file)
				
				gtk.gdk.threads_enter()
				self.parent.set_current_task('Scanning document list...')
				self.parent.spinner(False)
				gtk.gdk.threads_leave()				
			else:
				new_file = old_file
			
			p.append(new_file)
			it = self.main.ui.ls_documents.iter_next(it)
		
		gtk.gdk.threads_enter()
		self.parent.spinner(False)
		gtk.gdk.threads_leave()
		
		self.parent.set_max(1. / (len(p) - 1))
		os.rename(self.catn(p, tmp), self.main.ui.txt_cat_output.get_text())
		
		Session.temp_delete('cat')
		
		gtk.gdk.threads_enter()
		self.parent.set_current_task("Done.")
		gtk.gdk.threads_leave()
		
		self.parent.response(gtk.RESPONSE_OK)
		
class DirectoryCrawlerThread(ProgressThread):
	def __init__(self, parent, data):
		self.main = data[0]
		self.d = data[1]
		self.isrec = data[2]
		ProgressThread.__init__(self, parent)
			   
	def run(self): ## TODO: Implement the Cancel button
		i = 1
		if self.isrec: ## Scan folders recursively
			gtk.gdk.threads_enter()
			self.parent.set_current_task("Scanning directory tree... This may take some time...")
			gtk.gdk.threads_leave()
			toadd = []

			gtk.gdk.threads_enter()
			self.parent.spinner(True)
			gtk.gdk.threads_leave()

			for root, sdirs, files in os.walk(self.d, followlinks=True):
				for f in files:
					toadd.append("%s/%s" % (root, f))
				if self.has_stopped(): return

			gtk.gdk.threads_enter()
			self.parent.spinner(False)
			gtk.gdk.threads_leave()
			
			tot = float(len(toadd))
			self.parent.set_current_task("A total of %d files have been found. Looking for supported documents..." % len(toadd))
			for f in toadd:
				gtk.gdk.threads_enter()
				self.parent.set_fraction(i / tot)
				self.parent.set_current_task('Checking the MIME Type of "%s"' % f)
				self.main.add_file(f)
				gtk.gdk.threads_leave()
				i += 1
				if self.has_stopped(): return

		else: ## Just this folder
			dirs = os.listdir(self.d)
			tot = float(len(dirs))
			gtk.gdk.threads_enter()
			self.parent.set_current_task("A total of %d entries in the selected folder have been found. Looking for supported documents..." % len(dirs))
			gtk.gdk.threads_leave()
			for f in dirs:
				if os.path.isfile("%s/%s" % (self.d, f)):
					gtk.gdk.threads_enter()
					self.parent.set_fraction(i / tot)
					self.parent.set_current_task('Checking the MIME Type of "%s"' % f)
					self.main.add_file("%s/%s" % (self.d, f))
					gtk.gdk.threads_leave()
				i += 1
				if self.has_stopped(): return

		gtk.gdk.threads_enter()
		self.parent.set_current_task("Done.")
		gtk.gdk.threads_leave()
		self.parent.response(gtk.RESPONSE_OK)


class FileListThread(ProgressThread):
	def __init__(self, parent, data):
		self.main = data[0]
		self.flist = data[1]
		ProgressThread.__init__(self, parent)
			   
	def run(self):
		i = 1
		gtk.gdk.threads_enter()
		self.parent.set_current_task("A total of %d have been selected. Looking for supported documents..." % len(self.flist))
		gtk.gdk.threads_leave()
		tot = float(len(self.flist))
		for f in self.flist:
			self.parent.set_fraction(i / tot)
			gtk.gdk.threads_enter()
			self.parent.set_current_task('Checking the MIME Type of "%s"' % f)
			gtk.gdk.threads_leave()
			self.main.add_file(f)
			i += 1
			if self.has_stopped(): return

		gtk.gdk.threads_enter()
		self.parent.set_current_task("Done.")
		gtk.gdk.threads_leave()
		self.parent.response(gtk.RESPONSE_OK)

### END THREAD CLASSES

### BEGIN COMMAND CLASSES

class AddDocs(Command):
	def __init__(self, prev):
		self._prev = prev
		super(AddDocs, self).__init__()
		
	def do(self):
		self.redo = self.undo
		
	def undo(self):
		Session.main.ui.tv_documents.set_model(self._prev)
		self._prev, Session.main.ui.ls_documents = Session.main.ui.ls_documents, self._prev

class SetDocInfo(Command):
	_widgets = ['author', 'creator', 'producer', 'subject', 'title']
	
	def do(self):
		self.redo = self.undo
		
		self._next = [getattr(Session.main.ui, 'txt_%s' % widget).get_text() for widget in self._widgets]
		self._prev = Session.main.get_from_selected_row(4, 5, 6, 7, 8)
		self._execute()
	
	def undo(self):
		self._prev, self._next = self._next, self._prev
		self._execute()
		
	def _execute(self):
		i = 4
		for field in self._next:
			Session.main.set_at_selected_row(i, field)
			i += 1
			

### END COMMAND CLASSES

# See gtumbler_lib.Window.py for more details about how this class works
class GtumblerWindow(Window):
	__gtype_name__ = "GtumblerWindow"
	
	def finish_initializing(self, builder): # pylint: disable=E1002
		"""Set up the main window"""
		super(GtumblerWindow, self).finish_initializing(builder)

		self.AboutDialog = AboutGtumblerDialog
		self.PreferencesDialog = PreferencesGtumblerDialog

		# Code for other initialization actions should be added here.
		self.ui.fc_add_folder.add_buttons(gtk.STOCK_OK, gtk.RESPONSE_OK,
										  gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)

### BEGIN LOCAL HELPERS

	def add_file(self, f):
		# Trust the MIME Type instead of the extension (which might be
		# missing)
		mime = shell.get_mimetype(f)
		
		# Test for a valid MIME Type
		if mime not in mimes.CONVERT_MIMES:
			if mime not in mimes.CAT_MIMES:
				return
			else:
				with file(f, 'rb') as fin:
					pdf_in = pdf.PdfFileReader(fin)
					doc_info = pdf_in.getDocumentInfo()
				author, creator, producer, subject, title = [doc_info.author, doc_info.creator, doc_info.producer, doc_info.subject, doc_info.title]
		else:
			author, creator, producer, subject, title = [None] * 5
			
		h, t = os.path.split(f)
		# FIELDS -- UID (I) | filename (CA) | path (CA) | mimetype (CA) | author (CA) | creator (CA) | producer (CA) | subject (CA) | title (CA)
		Session.main.ui.ls_documents.append([Session.assign_uid(), # UID (I)
                                             t,                    # filename (CA)
                                             h,                    # path (CA)
                                             mime,				   # mimetype (CA)
                                             author,			   # author (CA)
                                             creator,              # creator (CA)
                                             producer,             # producer (CA)
                                             subject,              # subject (CA)
                                             title])               # title (CA)		
	
	def concatenate(self):
		pd = ProgressDialog()
		pd.thread = ConcatenateThread(pd, self)
		resp = pd.start()
		if resp != gtk.RESPONSE_OK:
			pd.stop()
		pd.destroy()
	
	def duplicate_doc_list(self):
		new_list = gtk.ListStore(gobject.TYPE_INT, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING)
		
		it = self.ui.ls_documents.get_iter_first()
		while it:
			new_list.append(self.ui.ls_documents.get(it, 0, 1, 2, 3, 4, 5, 6, 7, 8))
			it = self.ui.ls_documents.iter_next(it)
			
		return new_list
	
	def get_from_selected_row(self, *args):
		model, it = self.ui.tv_documents.get_selection().get_selected()
		return model.get(it, *args) if it else None
	
	def process(self):
		pass
		
	def set_at_selected_row(self, *args):
		model, it = self.ui.tv_documents.get_selection().get_selected()
		return model.set_value(it, *args) if it else None
		
### END LOCAL HELPERS

### BEGIN SIGNAL HANDLERS

	def on_btn_cat_browse_clicked(self, widget, data = None):
		out_dlg = gtk.FileChooserDialog("Specify the file name of the output document", self,
										 buttons = (gtk.STOCK_OK, gtk.RESPONSE_OK,
													gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL),
										 action = gtk.FILE_CHOOSER_ACTION_SAVE)
		
		resp = out_dlg.run()
		if resp == gtk.RESPONSE_OK:
			self.ui.txt_cat_output.set_text(out_dlg.get_filename())
		out_dlg.destroy()

	def on_btn_doc_info_apply_clicked(self, widget, data = None):
		SetDocInfo()

	def on_btn_pattern_help_toggled(self, widget, data = None):

		if widget.get_active():
			rect = txt = widget.get_allocation()
			x, y = widget.translate_coordinates(widget.get_toplevel(), 0, 0)
			wx, wy = self.get_position()
			w, h = self.ui.win_pattern.get_size()

			self.ui.win_pattern.move(wx + x - w + rect.width, wy - 6)
			
			self.ui.win_pattern.show()
		else:
			self.ui.win_pattern.hide()

	def on_btn_perform_action_clicked(self, widget, data = None):
		# Determine the action
		if self.ui.rdo_cat.get_active():  # Concatenate
			self.concatenate()
		else:							 # Process separately
			self.process()

	def on_ls_documents_row_changed(self, widget, path, it, *args):
		print "gotcha"
		self.on_tv_documents_cursor_changed(widget, *args)

	def on_rdo_cat_toggled(self, widget, data = None):
		self.ui.hb_cat.set_sensitive(widget.get_active())
		self.ui.vb_separate.set_sensitive(not widget.get_active())

	def on_rdo_overwrite_toggled(self, widget, data = None):
		self.ui.vb_saveas.set_sensitive(not widget.get_active())


	def on_tb_add_clicked(self, widget, data = None):
		##TODO: Figure out what the hell is wrong with this procedure!!!!
		open_dialog = gtk.FileChooserDialog("Select a document", self,
											buttons = (gtk.STOCK_OK, gtk.RESPONSE_OK,
													   gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
		open_dialog.set_select_multiple(True)
		
		resp = open_dialog.run()
		pd = None
		if resp == gtk.RESPONSE_OK:
			files = open_dialog.get_filenames()
			open_dialog.hide()
			prev = self.duplicate_doc_list()
			with ProgressDialog() as pd:
				pd.thread = FileListThread(pd, [self, files])
				resp = pd.start()
				if resp != gtk.RESPONSE_OK:
					pd.stop()
				else:
					AddDocs(prev)
		else:
			open_dialog.hide()
	
	def on_tb_add_folder_activate(self, widget, data = None):
		resp = self.ui.fc_add_folder.run()
		if resp == gtk.RESPONSE_OK:
			d = self.ui.fc_add_folder.get_filename()
			self.ui.fc_add_folder.hide()
			prev = self.duplicate_doc_list()
			with ProgressDialog() as pd:
				pd.thread = DirectoryCrawlerThread(pd, [self, d, self.ui.chk_recursive.get_active()])
				resp = pd.start()
				if resp != gtk.RESPONSE_OK:
					pd.stop()
				else:
					AddDocs(prev)
		else:
			self.ui.fc_add_folder.hide()

	def on_tb_redo_clicked(self, widget, data = None):
		Session.redo()
		
	def on_tb_undo_clicked(self, widget, data = None):
		Session.undo()
		
	def on_tv_documents_cursor_changed(self, widget, data = None):
		info = self.get_from_selected_row(4, 5, 6, 7, 8)
		widgets = [self.ui.txt_author, self.ui.txt_creator, self.ui.txt_producer, self.ui.txt_subject, self.ui.txt_title]
		i = 0
		for i in range(len(info)): widgets[i].set_text(info[i] if info[i] else '')
		
### END SIGNAL HANDLERS
